
#define __KERNEL__
#define MODULE
#define __KERNEL_SYSCALLS__
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/fs.h>

#include <linux/types.h>
#include <linux/unistd.h>

#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>
#include <asm/msr.h>


#include "params.h"

int errno;

void ksys_print(char *name, unsigned long value)
{
    printk("ksystime: %3li -- %s\n", value, name);
}

int init_module(void)
{
    mm_segment_t fs;
    unsigned long ini, end, now, best, tsc;
    int i;
    char buffer[40];
    struct file *file;
    ssize_t (*readptr)(struct file *, char *, size_t, loff_t *);
    loff_t *posptr;

#define measure_time(code) \
    for (i = 0; i < NTRIALS; i++) { \
	rdtscl(ini); \
        code; \
	rdtscl(end); \
	now = end - ini; \
	if (now < best) best = now; \
    }

    /* time rdtsc (i.e. no code) */
    best = ~0;
    measure_time( 0 );
    tsc = best;
    ksys_print("tsc", tsc);

    /* prepare to invoke a system call */
    fs = get_fs();
    set_fs (get_ds());

    /* time an empty read() */
    best = ~0;
    measure_time( read(0 /* stdin */, buffer, 40) );
    ksys_print("read()", best - tsc);

    /* call sys_read directly. This is disabled because
       you need to know the exact address of sys_read in kernel
       space. The value shown is the one I found in the system
       map of my current kernel. Please be careful */
    {
	ssize_t (*sys_read_ptr)(unsigned int , char *, size_t);
	sys_read_ptr = 0xc012aa50; /* THIS MUST BE CORRECTED */

	best = ~0;
	measure_time( (*sys_read_ptr)(0 /* stdin */, buffer, 40) );
	ksys_print("sys_read()", best - tsc);
    }

    /* use the file operation directly */
    file = fget(0 /* fd */);
    if (file && file->f_op && file->f_op->read) {
	best = ~0;
	measure_time(
	    file->f_op->read(file, buffer, 40, &file->f_pos)
	);
	ksys_print("f_op->read()", best - tsc);
    }

    /* cache the read ptr */
    if (file && file->f_op && file->f_op->read) {
	readptr = file->f_op->read;
	posptr = &file->f_pos;
	best = ~0;
	measure_time(
	    readptr(file, buffer, 40, posptr)
	);
	ksys_print("cached_f_op_read()", best - tsc);
    }
	
    if (file) fput(file);

    /* restore fs and make insmod fail */
    set_fs (fs);
    return -EINVAL;
}


void cleanup_module(void) {} /* never used */
